@@ -1,8 +1,9 @@
 #!/usr/bin/env python
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
+from collections.abc import Iterable, Iterator
 from gettext import gettext as _
-from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Iterator, List, Optional
+from typing import TYPE_CHECKING, Any, Callable, Optional
 
 from .constants import is_macos
 from .fast_data_types import (
@@ -37,7 +38,7 @@ def keyboard_mode_name(screen: ScreenType) -> str:
     return 'application' if screen.cursor_key_mode else 'normal'
 
 
-def get_shortcut(keymap: KeyMap, ev: KeyEvent) -> Optional[List[KeyDefinition]]:
+def get_shortcut(keymap: KeyMap, ev: KeyEvent) -> Optional[list[KeyDefinition]]:
     mods = ev.mods & mod_mask
     ans = keymap.get(SingleKey(mods, False, ev.key))
     if ans is None and ev.shifted_key and mods & GLFW_MOD_SHIFT:
@@ -63,8 +64,8 @@ class Mappings:
 
     ' Manage all keyboard mappings '
 
-    def __init__(self, global_shortcuts:Optional[Dict[str, SingleKey]] = None, callback_on_mode_change: Callable[[], Any] = lambda: None) -> None:
-        self.keyboard_mode_stack: List[KeyboardMode] = []
+    def __init__(self, global_shortcuts:Optional[dict[str, SingleKey]] = None, callback_on_mode_change: Callable[[], Any] = lambda: None) -> None:
+        self.keyboard_mode_stack: list[KeyboardMode] = []
         self.update_keymap(global_shortcuts)
         self.callback_on_mode_change = callback_on_mode_change
 
@@ -72,7 +73,7 @@ class Mappings:
     def current_keyboard_mode_name(self) -> str:
         return self.keyboard_mode_stack[-1].name if self.keyboard_mode_stack else ''
 
-    def update_keymap(self, global_shortcuts:Optional[Dict[str, SingleKey]] = None) -> None:
+    def update_keymap(self, global_shortcuts: Optional[dict[str, SingleKey]] = None) -> None:
         if global_shortcuts is None:
             global_shortcuts = self.set_cocoa_global_shortcuts(self.get_options()) if is_macos else {}
         self.global_shortcuts_map: KeyMap = {v: [KeyDefinition(definition=k)] for k, v in global_shortcuts.items()}
@@ -114,7 +115,7 @@ class Mappings:
         mode = self.keyboard_modes[new_mode]
         self._push_keyboard_mode(mode)
 
-    def matching_key_actions(self, candidates: Iterable[KeyDefinition]) -> List[KeyDefinition]:
+    def matching_key_actions(self, candidates: Iterable[KeyDefinition]) -> list[KeyDefinition]:
         w = self.get_active_window()
         matches = []
         has_sequence_match = False
@@ -244,7 +245,7 @@ class Mappings:
         if b.args.debug_keyboard:
             print(*args, end=end, flush=True)
 
-    def set_cocoa_global_shortcuts(self, opts: Options) -> Dict[str, SingleKey]:
+    def set_cocoa_global_shortcuts(self, opts: Options) -> dict[str, SingleKey]:
         from .main import set_cocoa_global_shortcuts
         return set_cocoa_global_shortcuts(opts)
     # }}}
